/***************************************************************************
                          sanetracker.cpp  -  description
                             -------------------
    begin                : Sat Jun  8 15:33:59 CEST 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for QT
#include <qdir.h>
#include <qpainter.h>

// include files for KDE
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <klocale.h>
#include <kstdaction.h>
#include <knuminput.h>

// application specific includes
#include "sanetracker.h"
#include "sanetrackerview.h"
#include "sanetrackerdoc.h"
#include "midi.h"
#include "mididriver.h"
#include "mixer.h"
#include "scheduler.h"
#include "pattern.h"
#include "piano.h"
#include "configuration.h"

#define ID_STATUS_MSG 1

#include <kdebug.h>

SanetrackerApp::SanetrackerApp(QWidget*, const char* name)
  : KMainWindow(0, name),
    configuration(this),
    mixer(0)
{
  MIDIDriver::init();

  statusBar()->insertItem(i18n("Ready."), ID_STATUS_MSG);
  
  initActions();
  
  configuration.load();

  MIDIDriver::reconfigure(&configuration);

  scheduler = new Scheduler(this, configuration);

  initDocument();
  initView();

  octave = 4;
  
  connect(this, SIGNAL(keyPressed(uint, uint)), view->pattern, SLOT(keyPressed(uint, uint)));
  connect(this, SIGNAL(keyReleased(uint, uint)), view->pattern, SLOT(keyReleased(uint, uint)));
  connect(this, SIGNAL(keyPressed(uint, uint)), view->piano, SLOT(keyPressed(uint)));
  connect(this, SIGNAL(keyReleased(uint, uint)), view->piano, SLOT(keyReleased(uint)));

  // disable actions at startup
  fileSave->setEnabled(false);
  editCut->setEnabled(false);
  editCopy->setEnabled(false);
  editPaste->setEnabled(false);

  if(!configuration.portCount())
  {
    configuration.autoconfig();

    configuration.configure(1, true);
  }

  scheduler->start();
}

SanetrackerApp::~SanetrackerApp()
{
  scheduler->stop();
}

bool SanetrackerApp::event(QEvent* event)
{
  if(event->type() == QEvent::User) // MIDI event
  {
    MIDIMessageEvent* midiEvent = (MIDIMessageEvent*) event;

    switch(midiEvent->status() & 0xF0)
    {
    case MIDI_NOTE_OFF(0):

      emit keyReleased(midiEvent->data(0), midiEvent->data(1));
        
      break;

    case MIDI_NOTE_ON(0):

      if(midiEvent->data(1) == 0)
        emit keyReleased(midiEvent->data(0), 127);
      else
        emit keyPressed(midiEvent->data(0), midiEvent->data(1));
        
      break;
    }
    
    return true;
  }
  else if(event->type() == QEvent::User + 1) // Position event
  {
    view->positionChanged();

    return true;
  }
  else if(event->type() == QEvent::KeyPress || event->type() == QEvent::KeyRelease)
  {
    if(QWidget::event(event))
      return true;

    QKeyEvent* keyEvent = (QKeyEvent*) event;
      
    if(keyEvent->isAutoRepeat())
      return true;

    static const char keyboard[24] =
    {// C   C#  D   D#  E   F   F#  G   G#  A   A#  B
       'Z','S','X','D','C','V','G','B','H','N','J','M', // octave + 0
       'Q','2','W','3','E','R','5','T','6','Y','7','U', // octave + 1
    };

    for(uint i = 0; i < 24; ++i)
    {
      if(keyEvent->key() == keyboard[i])
      {
        if(event->type() == QEvent::KeyPress)
          emit keyPressed(i + octave * 12, 127);
        else
          emit keyReleased(i + octave * 12, 127);
        
        return true;
      }
    }
    
    switch(keyEvent->key())
    {
      case Key_F1: octave = 0; break;
      case Key_F2: octave = 1; break;
      case Key_F3: octave = 2; break;
      case Key_F4: octave = 3; break;
      case Key_F5: octave = 4; break;
      case Key_F6: octave = 5; break;
      case Key_F7: octave = 6; break;
      case Key_F8: octave = 7; break;
      case Key_F9: octave = 8; break;
      case Key_F10: octave = 9; break;
      case Key_F11: octave = 10; break;
    }

    return true;
  }
  else if(event->type() == QEvent::Close)
  {
    slotFileQuit();

    return true;
  }
  else
  {
    return QWidget::event(event);
  }
}

void SanetrackerApp::initActions()
{
  fileNew = KStdAction::openNew(this, SLOT(slotFileNew()), actionCollection());
  fileOpen = KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
  fileOpenRecent
    = KStdAction::openRecent(this, SLOT(slotFileOpenRecent(const KURL&)),
                             actionCollection());
  fileSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
  fileSaveAs = KStdAction::saveAs(this, SLOT(slotFileSaveAs()),
                                  actionCollection());
  fileQuit = KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());
  editCut = KStdAction::cut(this, SLOT(slotEditCut()), actionCollection());
  editCopy = KStdAction::copy(this, SLOT(slotEditCopy()), actionCollection());
  editPaste = KStdAction::paste(this, SLOT(slotEditPaste()), actionCollection());
  viewToolBar = KStdAction::showToolbar(this, SLOT(slotViewToolBar()),
                                        actionCollection());
  viewStatusBar = KStdAction::showStatusbar(this, SLOT(slotViewStatusBar()),
                                            actionCollection());
  settingsConfigure = new KAction(i18n("&Configure Sanetracker..."), 0, 0,
                                  this, SLOT(slotSettingsConfigure()),
                                  actionCollection(), "configure");
  openMixer = new KAction(i18n("Show &Mixer Window..."), CTRL + Key_M, this,
                          SLOT(slotOpenMixer()), actionCollection(),
                          "showmixer");

  fileNew->setStatusText(i18n("Creates a new document"));
  fileOpen->setStatusText(i18n("Opens an existing document"));
  fileOpenRecent->setStatusText(i18n("Opens a recently used file"));
  fileSave->setStatusText(i18n("Saves the actual document"));
  fileSaveAs->setStatusText(i18n("Saves the actual document as..."));
  fileQuit->setStatusText(i18n("Quits the application"));
  editCut->setStatusText(i18n("Cuts the selected section and puts it to the clipboard"));
  editCopy->setStatusText(i18n("Copies the selected section to the clipboard"));
  editPaste->setStatusText(i18n("Pastes the clipboard contents to actual position"));
  viewToolBar->setStatusText(i18n("Enables/disables the toolbar"));
  viewStatusBar->setStatusText(i18n("Enables/disables the statusbar"));
  settingsConfigure->setStatusText(i18n("Opens a dialog for configuring Sanetracker"));
  openMixer->setStatusText(i18n("Shows the mixer window"));

  createGUI("sanetrackerui.rc");
}

void SanetrackerApp::initDocument()
{
  doc = new SanetrackerDoc(this, configuration);
  doc->newDocument();

  connect(doc, SIGNAL(modified(bool)), this, SLOT(slotModified(bool)));
}

void SanetrackerApp::initView()
{ 
  view = new SanetrackerView(this, &configuration, *scheduler);
  setCentralWidget(view);	
  setCaption(doc->URL().fileName(), false);
}

void SanetrackerApp::openDocumentFile(const KURL& url)
{
  slotStatusMsg(i18n("Opening file..."));

  doc->openDocument(url);
  fileOpenRecent->addURL(url);

  slotStatusMsg(i18n("Ready."));
}


SanetrackerDoc* SanetrackerApp::getDocument() const
{
  return doc;
}

SanetrackerView* SanetrackerApp::getView() const
{
  return view;
}

void SanetrackerApp::saveProperties(KConfig *_cfg)
{
  if(doc->URL().fileName() != i18n("Untitled") && !doc->isModified())
  {
    // saving to tempfile not necessary

  }
  else
  {
    KURL url=doc->URL();	
    _cfg->writeEntry("filename", url.url());
    _cfg->writeEntry("modified", doc->isModified());
    QString tempname = kapp->tempSaveName(url.url());
    QString tempurl= KURL::encode_string(tempname);
    KURL _url(tempurl);
    doc->saveDocument(_url);
  }
}


void SanetrackerApp::readProperties(KConfig* _cfg)
{
  QString filename = _cfg->readEntry("filename", "");
  KURL url(filename);
  bool modified = _cfg->readBoolEntry("modified", false);
  if(modified)
  {
    bool canRecover;
    QString tempname = kapp->checkRecoverFile(filename, canRecover);
    KURL _url(tempname);
  	
    if(canRecover)
    {
      doc->openDocument(_url);
      doc->setModified();
      setCaption(_url.fileName(), true);
      QFile::remove(tempname);
    }
  }
  else
  {
    if(!filename.isEmpty())
    {
      doc->openDocument(url);
      setCaption(url.fileName(), false);
    }
  }
}

void SanetrackerApp::reconfigure()
{
  bool shouldRun = !scheduler->isStopped();

  if(scheduler->running())
    scheduler->stop();

  MIDIDriver::reconfigure(&configuration);
  doc->reconfigure();
  view->reconfigure();

  if(shouldRun)
    scheduler->start();
}

/////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATION
/////////////////////////////////////////////////////////////////////

void SanetrackerApp::slotSettingsConfigure()
{
  slotStatusMsg(i18n("Opening the configuration dialog..."));

  configuration.configure();

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileNew()
{
  slotStatusMsg(i18n("Creating new document..."));

  if(saveModified())
  {
    doc->newDocument();
    setCaption(doc->URL().fileName(), false);

    if(mixer)
    {
      delete mixer;

      mixer = 0;
    }
  }

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileOpen()
{
  slotStatusMsg(i18n("Opening file..."));

  if(saveModified())
  {
    KURL url = KFileDialog::getOpenURL(QString::null,
        i18n("*.mid|MIDI Files\n*|All files"), this, i18n("Open File..."));

    if(!url.isEmpty())
    {
      doc->openDocument(url);
      setCaption(url.fileName(), false);
      fileOpenRecent->addURL( url );

      if(mixer)
      {
        delete mixer;

        mixer = 0;
      }
    }
  }

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileOpenRecent(const KURL& url)
{
  slotStatusMsg(i18n("Opening file..."));

  if(saveModified())
  {
    doc->openDocument(url);
    setCaption(url.fileName(), false);

    if(mixer)
    {
      delete mixer;

      mixer = 0;
    }
  }

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileSave()
{
  slotStatusMsg(i18n("Saving file..."));

  if(doc->URL().fileName() == i18n("Untitled"))
  {
    slotFileSaveAs();
  }
  else
  {
    if(!doc->saveDocument(doc->URL()))
    {
      slotStatusMsg(i18n("Save operation failed."));

      return;
    }
  }

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileSaveAs()
{
  slotStatusMsg(i18n("Saving file with a new filename..."));

  KURL url = KFileDialog::getSaveURL(QDir::currentDirPath(),
        i18n("*.mid|MIDI Files\n*|All files"), this, i18n("Save as..."));

  if(!url.isEmpty())
  {
    if(!doc->saveDocument(url))
    {
      slotStatusMsg(i18n("Save operation failed."));

      return;
    }

    fileOpenRecent->addURL(url);
    setCaption(url.fileName(), doc->isModified());
  }

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotFileQuit()
{
  slotStatusMsg(i18n("Exiting..."));

  if(saveModified())
  {
    MIDIDriver::close();

    scheduler->stop();

    configuration.save();

    kapp->quit();
  }
  else
  {
    slotStatusMsg(i18n("Ready."));
  }
}

void SanetrackerApp::slotEditCut()
{
  slotStatusMsg(i18n("Cutting selection..."));

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotEditCopy()
{
  slotStatusMsg(i18n("Copying selection to clipboard..."));

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotEditPaste()
{
  slotStatusMsg(i18n("Inserting clipboard contents..."));

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotViewToolBar()
{
  slotStatusMsg(i18n("Toggling toolbar..."));

  if(!viewToolBar->isChecked())
    toolBar("mainToolBar")->hide();
  else
    toolBar("mainToolBar")->show();

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotViewStatusBar()
{
  slotStatusMsg(i18n("Toggle the statusbar..."));

  if(!viewStatusBar->isChecked())
    statusBar()->hide();
  else
    statusBar()->show();

  slotStatusMsg(i18n("Ready."));
}

void SanetrackerApp::slotStatusMsg(const QString &text)
{
  statusBar()->clear();
  statusBar()->changeItem(text, ID_STATUS_MSG);
}

void SanetrackerApp::slotModified(bool modified)
{
  fileSave->setEnabled(modified);
}

void SanetrackerApp::slotOpenMixer()
{
  if(!mixer)
    mixer = new Mixer(*doc, configuration);

  if(mixer->isVisible())
    mixer->raise();
  else
    mixer->show();
}

bool SanetrackerApp::saveModified()
{
  bool completed = true;

  if(doc->isModified())
  {
    int want_save = KMessageBox::warningYesNoCancel(this,
      i18n("The current file has been modified.\n"
           "Do you want to save it?"),
      i18n("Warning"));

    switch(want_save)
    {
    case KMessageBox::Yes:

      slotFileSave();

      break;

    case KMessageBox::No:

      doc->setModified(false);

      break;

    case KMessageBox::Cancel:

      completed = false;

      break;

    default:

      completed = false;

      break;
    }
  }

  return completed;
}

